# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Base class for resources based on DB models. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
from logging import getLogger
from typing import Text

# 3rd party imports
import jsonschema
from serafin import Fieldspec, serialize

# local imports
from .resource import RestResource
from .util import iter_public_props


L = getLogger(__name__)


class ModelResource(RestResource):
    """ Base class for resources based on DB models. """

    model = None
    spec = Fieldspec('*')
    schema = {}
    read_only = []

    class AlreadyExists(RuntimeError):
        """ Raised when an object already exists. """
        pass

    class ValidationError(RuntimeError):
        """ Raised by .validate() if it fails. """
        def __init__(self, jsonschema_error):
            self.detail = jsonschema_error
            super(ModelResource.ValidationError, self).__init__(
                str(jsonschema_error)
            )

    def validate(self, data, schema=None):
        """ Validate the *data* according to the given *schema*.

        :param Dict[str, Any] data:
            A dictionary of data. Probably coming from the user in some way.
        :param Dict[str, Any] schema:
            JSONSchema describing the required data structure.
        :raises ModelResource.ValidationError:
            If the validation fails. No value is returned.
        """
        try:
            jsonschema.validate(data, schema or self.schema)
        except jsonschema.ValidationError as ex:
            raise ModelResource.ValidationError(ex)

    def serialize(self, item_or_items, spec=None):
        """ Serialize an item or items into a dict.

        This will just call serafin.serialize using the model spec (defined
        class wide in the model - '*' by default).

        :param item_or_items:
        :return Dict[Any, Any]:
            A dict with python native content. Can be easily dumped to any
            format like JSON or YAML.
        """
        if spec is None:
            spec = self.spec

        return serialize(item_or_items, spec)

    def deserialize(self, data):
        """ Convert JSON data into model field types.

        The value returned by this function can be used directly to create new
        item and update existing ones.
        """
        return data

    def implements(self, rest_verb):
        # type: (Text) -> bool
        """ Check whether this model resource implements a given REST verb.

        Args:
            rest_verb (str):
                The REST verb you want to check. Possible values are *create*,
                *query*, *get*, *update* and *delete*.

        Returns:
            bool: **True** if the given REST verb is implemented, **False**
                otherwise.
        """
        test = {
            'create': lambda: self.create_item(None, {}, {}),
            'query': lambda: self.query_items(None, {}, {}),
            'get': lambda: self.get_item(None, {}, {}),
            'update': lambda: self.update_item(None, {}, {}),
            'delete': lambda: self.delete_item(None, {}, {}),
        }.get(rest_verb)

        if test:
            try:
                test()
                return True
            except NotImplementedError:
                return False
            except:
                return True
        else:
            return False

    def item_for_request(self, request):
        """ Create new model item. """
        del request     # Unused here
        raise NotImplementedError(
            "All resources must  implement .item_for_request()"
        )

    @property
    def public_props(self):
        """ All public properties on the resource model. """
        if not hasattr(self, '_public_props'):
            self._public_props = [
                name for name, _ in iter_public_props(self.model)
            ]
        return self._public_props

    def create_item(self, request, params, payload):
        """ Create new model item. """
        raise NotImplementedError("Must implement .create_item()")

    def update_item(self, request, params, payload):
        """ Update existing model item. """
        raise NotImplementedError("Must implement .update_item()")

    def delete_item(self, request, params, payload):
        """ Delete model instance. """
        raise NotImplementedError("Must implement .delete_item()")

    def query_items(self, request, params, payload):
        """ Return a model query with the given filters.

        The query can be further customised like any ndb query.

        :return google.appengine.ext.ndb.Query:
            The query with the given filters already applied.
        """
        raise NotImplementedError("Must implement .query_items()")

    def get_item(self, request, params, payload):
        """ Get an item associated with the request.

        This is used by all detail views/actions to get the item that the
        request is concerned with (usually from the URL). This is an
        implementation detail and is highly dependant on the underlying web
        framework used.

        :param request:
            HTTP request.
        :return RestResource:
            The item associated with the request.
        """
        raise NotImplementedError("{}.get_item() not implemented".format(
            self.__class__.__name__
        ))

    def rest_query(self, request, params, payload):
        """ Query existing records as a list. """
        try:
            fields = params.pop('_fields', '*')

            filters = self.deserialize(params)
            items = self.query_items(request, filters, payload)

            spec = Fieldspec(self.spec).restrict(Fieldspec(fields))
            ret = self.serialize(items, spec)
            return 200, ret

        except NotImplementedError:
            return 404, {'detail': 'Not Found'}

    def rest_create(self, request, params, payload):
        """ Create a new record. """
        try:
            self.validate(payload, self.schema)

            values = self.deserialize(payload)
            item = self.create_item(request, params, values)

            return self.serialize(item)

        except ModelResource.ValidationError as ex:
            return 400, {'detail': str(ex)}

        except ModelResource.AlreadyExists:
            return 400, {'detail': 'Already exists'}

        except NotImplementedError:
            return 404, {'detail': 'Not Found'}

    def rest_get(self, request, params, payload):
        """ Get one record with the given id. """
        try:
            fields = Fieldspec(params.get('_fields', '*'))

            spec = Fieldspec(self.spec).restrict(fields)
            item = self.get_item(request, params, payload)

            if item is not None:
                return 200, self.serialize(item, spec)
            else:
                return 404, {'detail': "Not Found"}

        except NotImplementedError:
            return 404, {'detail': 'Not Found'}

    def rest_update(self, request, params, payload):
        """ Update existing item. """
        schema = {}
        schema.update(self.schema)

        if 'required' in schema:
            del schema['required']

        try:
            self.validate(payload, schema)

            values = self.deserialize(payload)
            read_only = (
                frozenset(self.read_only or []) | frozenset(self.public_props or [])
            )
            for name in read_only:
                values.pop(name, None)

            item = self.update_item(request, params, values)

            if item is not None:
                return 200, self.serialize(item)
            else:
                return 404, {'detail': "Not Found"}

        except ModelResource.ValidationError as ex:
            return 400, {'detail': str(ex)}

        except NotImplementedError:
            return 404, {'detail': 'Not Found'}

    def rest_delete(self, request, params, payload):
        """ DELETE detail. """
        try:
            self.delete_item(request, params, payload)

            return 204, {}

        except NotImplementedError:
            return 404, {'detail': 'Not Found'}


# Used only in type hint comments
del Text



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Rest endpoint base implementation.

Endpoints take care of interacting with HTTP. They for a bridge between HTTP
requests/responses and the REST resources. They should never be implemented
manually and ideally most apps should not need to subclass those but instead
implement all of their logic in the `RestResource` subclasses.

The code here is just an abstract base that holds the code common to all
endpoints. It is not self sufficient, but only minor web-framework dependent
code is left for the subclasses.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from collections import namedtuple
from logging import getLogger
from typing import List, Text

# 3rd party imports
import jsonschema

# local imports
from .resource import RestResource
from .actions import api_action
from . import url_params
from . import exc


L = getLogger(__name__)
RestResult = namedtuple('RestResult', 'status,headers,data')


class RawResponse(object):
    """ Wrapper around raw framework responses.

    This exists so it can help distinguish
    """
    def __init__(self, response):
        self.response = response


class RestEndpoint(object):
    """ Abstract base class for all REST endpoints.

    The endpoint responsibility is to provide a bridge between HTTP and
    `RestResource`.  The latter should never directly deal with HTTP.


    :param RestResource resource:
        `RestResource` instance. This supersedes whatever is passed in
        `res_cls`.
    :param type res_cls:
        Class derived from `RestResource`. If `resource` is not passed
        directly you can also specify the resource class. It will be used
        to instantiate the resource when the endpoint is created.
    :param bool protected:
        If set to *True*, all CRUD operations on the endpoint will require
        authorization.
    """

    def __init__(self, resource=None, res_cls=None, protected=False):
        resource = resource or getattr(self, 'resource', None)
        res_cls = res_cls or getattr(self, 'res_cls', None)

        if resource is not None and isinstance(resource, RestResource):
            self.resource = resource
        elif (
                res_cls is not None and
                isinstance(res_cls, type) and
                issubclass(res_cls, RestResource)
        ):
            self.resource = res_cls()
        else:
            raise ValueError(
                "You must specify the resource either as a static variable "
                "resource or res_cls or through corresponding constructor "
                "args"
            )

        self.protected = protected

    def is_resource(self, res_cls):
        """ Check if the given object is a resource class.

        :param type res_cls:
            Object to test.
        :return bool:
            **True** if *res_cls* is a resource class.
        """
        return not (isinstance(res_cls, type) and
                    issubclass(res_cls, RestResource))

    def get_ok_status(self, verb):
        """ Get a valid ok status code for a given rest verb. """
        if verb == 'create':
            return 201
        elif verb == 'delete':
            return 204
        else:
            return 200

    def authorize(self, request):
        """ Authorize user from request.

        By default it returns None, so no user. Treat it as unauthorized

        :param request:
            HTTP request.
        :return:
            A user instance (depending on the underlying storage/models) or None
            if not authorized.
        """
        return None

    def process_result(self, result, default_status):
        """ Unify the handler result into a stable format.

        The unified format is ``(status, headers, data)``.

        :param tuple|dict|list|RawResponse result:
            The result as returned from the handler
        :param int default_status:
            Default HTTP status to use (if result contains none).
        :return tuple(int, dict, dict):
            Tuple . The handlers can return data in
            the shortened forms, either ``(status, data)`` or just ``data``.
            This method will expand those results with default and return a
            unified result tuple.
        """
        status = default_status
        headers = {}
        data = None

        if isinstance(result, tuple):
            if len(result) == 2:
                status, data = result
            elif len(result) == 3:
                status, headers, data = result
            else:
                raise ValueError("REST result has too many items")

        elif isinstance(result, RawResponse):
            return result

        else:
            data = result

        return RestResult(status, headers, data)

    def call_rest_handler(self, method, request):
        """

        :param str method:
            HTTP method. Must be uppercase
        :param request:
            HTTP request. It is assumed the request object hast the following
            members: ``rest_keys``,
        :return tuple(int, dict, dict):
            Tuple ``(status, headers, data)``. The handlers can return data in
            the shortened forms, either ``(status, data)`` or just ``data``.
            This method will expand those results with default and return a
            unified result tuple.
        """
        my_pk = self.resource.get_pk(request)

        request.user = self.authorize(request)
        if self.protected and request.user is None:
            return RestResult(401, {}, {'detail': "Not Authorized"})

        rest_verb = self.determine_rest_verb(method, my_pk)
        if rest_verb is None:
            return RestResult(405, {}, None)

        handler = getattr(self.resource, 'rest_' + rest_verb, None)
        if handler is None or not callable(handler):
            return RestResult(405, {}, None)

        qs = self.extract_request_query_string(request)
        params = url_params.parse(qs)
        payload = self.extract_request_data(request)

        try:
            result = handler(request, params, payload)
            return self.process_result(result, self.get_ok_status(rest_verb))

        except exc.Error as ex:
            return RestResult(ex.status, {}, {'detail': ex.detail})

        except ValueError as ex:
            L.exception('Invalid REST invocation')
            L.error(ex)
            return RestResult(400, {}, {'detail': str(ex)})

        except NotImplementedError as ex:
            return RestResult(405, {}, {'detail': str(ex)})

        except Exception as ex:
            L.exception('Unhandled resource exception')
            L.error(ex)
            return RestResult(500, {}, {'detail': str(ex)})

    def call_action_handler(self, method, request, name, is_generic):
        """ Call an action on the bound resource.

        :param request:
            HTTP request that called the action.
        :param str name:
            The action name.
        :param bool is_generic:
            True if we should call a generic actions. Generic actions are
            executed globally for the resource. Non-generic actions are called
            on one specific resource instance.

        :return RestResult:
            RestResult tuple with the result of the REST call. This can be
            easily converted to any underlying framework.
        """
        action = self.find_action(name, is_generic, method)
        if action is None:
            return RestResult(404, {}, {'detail': "{} has no action: {}".format(
                self.resource.name, name
            )})

        meta = api_action.get_meta(action)
        user = self.authorize(request)

        if meta.protected and user is None:
            return RestResult(401, {}, {'detail': "Not Authorized"})

        if method.lower() not in meta.methods:
            return RestResult(405, {}, {
                'detail': "Action {}.{} does not support method {}".format(
                    self.resource.__class__.__name__, meta.name, method.upper()
                )
            })

        request.user = user
        qs = self.extract_request_query_string(request)
        params = url_params.parse(qs)
        payload = self.extract_request_data(request)

        if meta.schema is not None:
            try:
                jsonschema.validate(payload, meta.schema)
            except jsonschema.ValidationError as ex:
                return RestResult(400, {}, {'detail': str(ex)})

        try:
            result = action(request, params, payload)
            return self.process_result(result, 200)

        except exc.Error as ex:
            return RestResult(ex.status, {}, {'detail': ex.detail})

        except NotImplementedError as ex:
            return RestResult(405, {}, {'detail': str(ex)})

        except Exception as ex:
            L.exception('Unhandled resource exception')
            L.error(ex)
            return RestResult(500, {}, {'detail': str(ex)})

    def find_action(self, name, generic, method='post'):
        """ Find API action by name and kind.

        :param str name:
            The action name.
        :param bool is_generic:
            True if we should call a generic actions. Generic actions are
            executed globally for the resource. Non-generic actions are called
            on one specific resource instance.
        :return Function:
            Action handler with it's associated metadata. THe metadata can
            be accessed using ``api_action.get_meta(action)``
        """
        method = method.lower()

        for action in self.resource.rest_actions():
            meta = api_action.get_meta(action)

            if (
                    meta.name == name and
                    meta.generic == generic and
                    method in meta.methods
            ):
                return action

        return None

    @classmethod
    def determine_rest_verb(cls, http_method, pk):
        """ Return the REST operation associated with the given HTTP method.

        This is part of the `RestEndpoint` class because this way it becomes
        customisable by subclasses.

        :param str|unicode http_method:
            The HTTP method used for the request.
        :param str|int pk:
            The primary key extracted from the URL template.
        :return str|unicode:
            The REST verb associated with that request. This is also the name
            of the method on `RestResource` that will be called by the endpoint.
        """
        if http_method == 'GET':
            return 'query' if not pk else 'get'
        elif http_method == 'POST' and not pk:
            return 'create'
        elif http_method == 'PUT' and pk:
            return 'update'
        elif http_method == 'DELETE' and pk:
            return 'delete'
        elif http_method == 'OPTIONS':
            return 'options'
        elif http_method == 'HEAD':
            return 'head'
        else:
            return None

    @classmethod
    def extract_request_data(cls, request):
        """ Extract request data as a python data structure

        This method is here to provide an easy way to implement new framework
        support. The way the request data is handled is framework dependant
        so each endpoint implementation for the given framework must provide
        and implementation of this function.

        :param Request request:
            Underlying framework dependant request.
        :return:
            A python data structure (dict,list,string, primitive type) that is
            the request data extract from the underlying
        """
        raise NotImplementedError(
            "RestEndpoint sublasses must implement .extract_request_data()"
        )

    @classmethod
    def extract_request_query_string(cls, request):
        """ Extract request params as a python dictionary.

        This method is here to provide an easy way to implement new framework
        support. The way the request params are handled is framework dependant
        so each endpoint implementation for the given framework must provide
        an implementation of this function.

        :param Request request:
            Underlying framework dependant request.
        :return:
            A python dict with ``{name: value}`` pairs for all request query
            string params
        """
        return request.GET


# Used only in type hint comments
del List, Text



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Resource actions helpers """
from __future__ import absolute_import, unicode_literals

# 3rd party imports
import attr


@attr.s
class ActionMeta(object):
    """ Action metadata.

    When you have a action method ``Res.my_action`` you can access the metadata
    with ``api_action.get_meta(Res.my_action)``.
    """
    name = attr.ib()
    generic = attr.ib()
    protected = attr.ib()
    methods = attr.ib()
    action = attr.ib()
    schema = attr.ib()


class api_action(object):
    """ Decorator for defining resource actions.

    Actions are extra HTTP endpoints that can be called with POST.
    """
    ACTION_ATTR = '__api_action__'

    def __init__(self, name=None, generic=False, protected=True, methods=None,
                 schema=None):
        self.name = name
        self.generic = generic
        self.protected = protected
        self.schema = schema
        self.methods = [m.lower() for m in (methods or ['post'])]

    def __call__(self, fn):
        """ Decorator. """
        setattr(fn, self.ACTION_ATTR, ActionMeta(
            name=self.name or fn.__name__,
            generic=self.generic,
            protected=self.protected,
            methods=self.methods,
            schema=self.schema,
            action=fn,
        ))
        return fn

    @classmethod
    def is_action(cls, obj):
        """ Return true if obj is an API action. """
        return hasattr(obj, cls.ACTION_ATTR)

    @classmethod
    def get_meta(cls, action):
        """ Get action metadata """
        return getattr(action, cls.ACTION_ATTR, {})



# -*- coding: utf-8 -*-
# Copyright 2017-2019 Mateusz Klos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
'''

Working with resources
======================

**restible** resources implementation is divided into two classes:

    - `RestResource`
    - `ModelResource`

The `RestResource` class is a very basic wrapper around a REST resource where
everything has to be implemented manually. This is the most basic usage of
resitble and allows the maximum extensibility. The downside is that it requires
writing quite a lot of repetitive code in order to get it working. Because of
that, **restible** provides another base class that inherits from
`RestResource`: the `ModelResource` class.

Most of the time the REST resource will have some corresponding model class.
`ModelResource` contains code for most common operations that you would do on a
resource backed by model. It does not manipulate any model classes directly so
it's not DB dependant in any way. It only contains the code that would probably
have to be written for any model based resource, no matter what underlying
library you use to persist your models. There are separate libraries that
provide integration with most common frameworks/libraries:

:Django: `restible-django <https://github.com/novopl/restible-django>`_.
:SQLAlchemy: `restible-sqlalchemy <https://github.com/novopl/restible-sqlalchemy>`_.
:AppEngine NDB: `restible-appengine <https://github.com/novopl/restible-appengine>`_.


Simple app example
==================

In this section we will create a simple AppEngine application that uses
**restible** to build the REST API. We've chosen AppEngine as it comes with
everything needed (HTTP and ORM layers) and is easy to deploy without any
devops. The free AppEngine account let's you host 10 of your apps so that also
helps.

.. code-block:: python
    :linenos:

    import restible
    import webapp2
    from google.appengine.ext import ndb
    from restible_appengine.util import ndb_query_from_values

    class BlogPost(ndb.Model):
        title = ndb.StringProperty()
        content = ndb.TextProperty()
        created_at = ndb.DateTimeProperty(auto_now_add=True)


    class BlogPostResource(ndb.ModelResource):
        name = 'post'
        model = BlogPost
        spec = serafin.Fieldspec('*')
        route_params = [{"name": "user_id"}]

        def create_item(self, request, filters):
            """ Get a list of resources. """
            return ndb_query_from_values(models.Resource, filters)


    wsgi_app = webapp2.WSGIApplication(
        routes=[

        ],
        debug=constants.DEBUG,
        config=CONFIG
    )

'''
from __future__ import absolute_import, unicode_literals

# stdlib imports
import inspect
import re
from typing import Text

from .actions import api_action


class RestResource(object):
    """ Represents the operations available on the given resource.

    This class should not handle HTTP directly. Rather than that there should
    be a separate implementation that maps HTTP layer onto resources. This way
    we can swap the HTTP layer without reimplementing the resource. This also
    follows SRP.

    If the given `RestResource` subclass doesn't have one of the REST methods
    implemented, the endpoint should return 405 Method not allowed (or 404?)
    """

    _NAME_RE = re.compile(r'^[a-zA-Z_][\w\d_]*$')
    name = None
    parent = None

    def __init__(self):
        self._validate_name(self.name)

    def rest_actions(self):
        """ Return all actions defined on the current resource.

        :return List[function]:
            List of action handlers. To get the metadata for a given *action*
            just use ``api_action.get_meta(action)``
        """
        ATTR = '_actions'

        if not hasattr(self, ATTR):
            actions = []
            for _, method in inspect.getmembers(self, inspect.ismethod):
                if api_action.is_action(method):
                    actions.append(method)

            setattr(self, ATTR, actions)

        return getattr(self, ATTR)

    def get_pk(self, request):
        """ Read the current resource type PK from the request.

        :param HttpRequest request:
            An HTTP request. Must contain the ``.rest_keys`` dictionary injected
            by `RestEndpoint.dispatch()` method.
        :return:
            Primary key that is associated with this resource type. This is
            extracted from ``.rest_keys`` based on the resource name.
        """
        if hasattr(request, 'rest_keys'):

            if hasattr(self, 'route_params'):
                if len(self.route_params) == 1:
                    return request.rest_keys.get(self.route_params[0]['name'])
                else:
                    params = (
                        request.rest_keys.get(route_param['name'])
                        for route_param in self.route_params
                    )
                    return [x for x in params if x]
            else:
                return request.rest_keys.get('{name}_pk'.format(name=self.name))

    def _validate_name(self, name):
        """ Validate the name is a valid resource name. """
        if name is None or not self._NAME_RE.match(name):
            raise ValueError(
                'REST endpoint name contain only letters, digits and '
                'underscores and cannot start with a digit'
            )

    def implements(self, rest_verb):
        # type: (Text) -> bool
        """ Check whether this resource implements a given REST verb.

        Args:
            rest_verb (str):
                The REST verb you want to check. Possible values are *create*,
                *query*, *get*, *update* and *delete*.

        Returns:
            bool: **True** if the given REST verb is implemented, **False**
                otherwise.
        """
        test = {
            'create': lambda: self.rest_create(None, {}),
            'query': lambda: self.rest_query(None, {}),
            'get': lambda: self.rest_get(None, {}),
            'update': lambda: self.rest_update(None, {}),
            'delete': lambda: self.rest_delete(None),
        }.get(rest_verb)

        if test:
            try:
                test()
                return True
            except NotImplementedError:
                return False
            except:
                return True
        else:
            return False

    def rest_query(self, request, params, payload):
        """ GET list. """
        raise NotImplementedError(".rest_query() not implemented".format(
            self.__class__.__name__
        ))

    def rest_get(self, request, params, payload):
        """ GET detail. """
        raise NotImplementedError("{}.rest_get() not implemented".format(
            self.__class__.__name__
        ))

    def rest_create(self, request, params, payload):
        """ POST list. """
        raise NotImplementedError("{}.rest_create() not implemented".format(
            self.__class__.__name__
        ))

    def rest_update(self, request, params, payload):
        """ PUT detail. """
        raise NotImplementedError("{}.rest_update() not implemented".format(
            self.__class__.__name__
        ))

    def rest_delete(self, request, params, payload):
        """ DELETE detail. """
        raise NotImplementedError("{}.rest_delete() not implemented".format(
            self.__class__.__name__
        ))

    def rest_options(self, request, params, payload):
        """ OPTIONS list/detail. """
        raise NotImplementedError("{}.rest_options() not implemented".format(
            self.__class__.__name__
        ))

    def rest_head(self, request, params, payload):
        """ OPTIONS list/detail. """
        raise NotImplementedError("{}.rest_head() not implemented".format(
            self.__class__.__name__
        ))


# Used only in type hint comments
del Text
